import pandas as pd
import unittest
import tempfile
import sys

import jollyjack as jj
import palletjack as pj
import pyarrow.parquet as pq
import pyarrow as pa
import numpy as np
import platform
import os
import itertools
import torch
from pyarrow import fs
from parameterized import parameterized

chunk_size = 3
n_row_groups = 2
n_columns = 5
n_rows = n_row_groups * chunk_size
current_dir = os.path.dirname(os.path.realpath(__file__))

supported_dtype_encodings = [
    (pa.float16(), "PLAIN"),
    (pa.float16(), "BYTE_STREAM_SPLIT"),
    (pa.float32(), "PLAIN"),
    (pa.float32(), "BYTE_STREAM_SPLIT"),
    (pa.float64(), "PLAIN"),
    (pa.float64(), "BYTE_STREAM_SPLIT"),
    (pa.int32(), "PLAIN"),
    (pa.int32(), "BYTE_STREAM_SPLIT"),
    (pa.int32(), "DELTA_BINARY_PACKED"),
    (pa.int64(), "PLAIN"),
    (pa.int64(), "BYTE_STREAM_SPLIT"),
    (pa.int64(), "DELTA_BINARY_PACKED"),
]

os_name = platform.system()

numpy_to_torch_dtype_dict = {
    np.bool: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
}


def get_table_with_nulls(n_rows, n_columns, data_type=pa.float32()):

    nullable_types = {
        pa.float32(): "Float32",
        pa.float16(): pa.float16().to_pandas_dtype(),
        pa.float64(): "Float64",
        pa.int16(): "Int16",
        pa.int32(): "Int32",
        pa.int64(): "Int64",
    }
    data = {}
    for i in range(n_columns):
        if pa.types.is_integer(data_type):
            data[f"column_{i}"] = pd.array(
                np.random.randint(-100, 100, size=n_rows),
                dtype=nullable_types[data_type],
            )
        else:
            data[f"column_{i}"] = pd.array(
                np.random.uniform(-100, 100, size=n_rows),
                dtype=nullable_types[data_type],
            )

    df = pd.DataFrame(data)
    df.iloc[0, 0] = None

    # Convert to PyArrow Table
    return pa.Table.from_pandas(df)


def get_table(n_rows, n_columns, data_type=pa.float32()):
    # Generate a random 2D array of floats using NumPy
    # Each column in the array represents a column in the final table
    data = np.random.uniform(-100, 100, size=(n_rows, n_columns)).astype(np.float32)

    # Convert the NumPy array to a list of PyArrow Arrays, one for each column
    pa_arrays = [
        pa.array(data[:, i]).cast(data_type, safe=False) for i in range(n_columns)
    ]
    schema = pa.schema([(f"column_{i}", data_type) for i in range(n_columns)])
    # Create a PyArrow Table from the Arrays
    return pa.Table.from_arrays(pa_arrays, schema=schema)


class TestJollyJack(unittest.TestCase):

    @parameterized.expand(
        itertools.product([False, True], [False, True], [False, True])
    )
    def test_read_entire_table(self, pre_buffer, use_threads, use_memory_map):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            pr = pq.ParquetReader()
            pr.open(path)

            # Create an array of zeros
            np_array1 = np.zeros((n_rows, n_columns), dtype="f", order="F")

            row_begin = 0
            row_end = 0

            for rg in range(n_row_groups):
                row_begin = row_end
                row_end = row_begin + pr.metadata.row_group(rg).num_rows
                subset_view = np_array1[row_begin:row_end, :]
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=subset_view,
                    row_group_indices=[rg],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            expected_data = pr.read_all()
            self.assertTrue(np.array_equal(np_array1, expected_data))

            np_array2 = np.zeros((n_rows, n_columns), dtype="f", order="F")
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array2,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            self.assertTrue(np.array_equal(np_array2, expected_data))
            pr.close()

    @parameterized.expand(
        itertools.product([False, True], [False, True], [False, True])
    )
    def test_read_with_palletjack(self, pre_buffer, use_threads, use_memory_map):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            index_path = path + ".index"
            pj.generate_metadata_index(path, index_path)

            # Create an array of zeros
            np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")

            row_begin = 0
            row_end = 0

            for rg in range(n_row_groups):
                column_indices = list(range(n_columns))
                metadata = pj.read_metadata(
                    index_path, row_groups=[rg], column_indices=column_indices
                )

                row_begin = row_end
                row_end = row_begin + metadata.num_rows
                subset_view = np_array[row_begin:row_end, :]
                jj.read_into_numpy(
                    source=path,
                    metadata=metadata,
                    np_array=subset_view,
                    row_group_indices=[0],
                    column_indices=column_indices,
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all()
            self.assertTrue(np.array_equal(np_array, expected_data))
            pr.close()

    @parameterized.expand(
        itertools.product([False, True], [False, True], [False, True])
    )
    def test_read_nonzero_column_offset(self, pre_buffer, use_threads, use_memory_map):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=chunk_size, n_columns=n_columns)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an array of zeros
            cols = 2
            offset = n_columns - cols
            np_array = np.zeros((chunk_size, cols), dtype="f", order="F")

            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=[0],
                column_indices=range(offset, offset + cols),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all(column_indices=range(offset, offset + cols))
            self.assertTrue(np.array_equal(np_array, expected_data))
            pr.close()

    @parameterized.expand(
        itertools.product([False, True], [False, True], [False, True])
    )
    def test_read_unsupported_column_types(
        self, pre_buffer, use_threads, use_memory_map
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(
                n_rows=chunk_size, n_columns=n_columns, data_type=pa.bool_()
            )
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an array of zerosx
            np_array = np.zeros((chunk_size, n_columns), dtype="f", order="F")

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(n_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"has unsupported data type: 0!" in str(context.exception),
                context.exception,
            )

    @parameterized.expand(itertools.product([pa.float16(), pa.float32(), pa.float64()]))
    def test_read_unsupported_encoding_dictionary(self, dtype):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=chunk_size, n_columns=n_columns, data_type=dtype)
            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=True)

            # Create an array of zerosx
            np_array = np.zeros(
                (chunk_size, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(n_columns),
                )

            self.assertTrue(
                f"due to unsupported_encoding=RLE_DICTIONARY!"
                in str(context.exception),
                context.exception,
            )

    def test_read_unsupported_encoding_delta_byte_array(self):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(
                n_rows=chunk_size, n_columns=n_columns, data_type=pa.float16()
            )
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                column_encoding="DELTA_BYTE_ARRAY",
            )

            # Create an array of zerosx
            np_array = np.zeros(
                (chunk_size, n_columns), dtype=pa.float16().to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(n_columns),
                )

            self.assertTrue(
                f"due to unsupported_encoding=DELTA_BYTE_ARRAY!"
                in str(context.exception),
                context.exception,
            )

    @parameterized.expand(
        itertools.product(
            [False, True], [False, True], [False, True], supported_dtype_encodings
        )
    )
    def test_read_dtype_numpy(
        self, pre_buffer, use_threads, use_memory_map, dtype_encoding
    ):

        for n_row_groups, n_columns, chunk_size in [
            (1, 1, 1),
            (2, 2, 1),
            (1, 1, 2),
            (1, 1, 10),
            (1, 1, 100),
            (1, 1, 1_000),
            (1, 1, 10_000),
            (1, 1, 100_000),
            (1, 1, 1_000_000),
            (1, 1, 10_000_000),
            (
                1,
                1,
                10_000_001,
            ),  # +1 to make sure it is not a result of multip,lication of a round number
        ]:

            dtype = dtype_encoding[0]
            encoding = dtype_encoding[1]

            with self.subTest(
                (
                    n_row_groups,
                    n_columns,
                    chunk_size,
                    dtype,
                    pre_buffer,
                    use_threads,
                    encoding,
                )
            ):
                n_rows = n_row_groups * chunk_size
                with tempfile.TemporaryDirectory() as tmpdirname:
                    path = os.path.join(tmpdirname, "my.parquet")
                    table = get_table(
                        n_rows=n_rows, n_columns=n_columns, data_type=dtype
                    )
                    use_dictionary = False
                    column_encoding = encoding
                    if encoding in ["RLE_DICTIONARY", "PLAIN_DICTIONARY"]:
                        use_dictionary = True
                        column_encoding = None

                    pq.write_table(
                        table,
                        path,
                        row_group_size=chunk_size,
                        use_dictionary=use_dictionary,
                        write_statistics=False,
                        store_schema=False,
                        column_encoding=column_encoding,
                    )

                    # Create an empty array
                    np_array = np.zeros(
                        (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
                    )

                    jj.read_into_numpy(
                        source=path,
                        metadata=None,
                        np_array=np_array,
                        row_group_indices=range(n_row_groups),
                        column_indices=range(n_columns),
                        pre_buffer=pre_buffer,
                        use_threads=use_threads,
                        use_memory_map=use_memory_map,
                    )

                    pr = pq.ParquetReader()
                    pr.open(path)
                    expected_data = pr.read_all().to_pandas().to_numpy()
                    self.assertTrue(
                        np.array_equal(np_array, expected_data),
                        f"{np_array}\n{expected_data}",
                    )
                    pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True], [False, True], [False, True], supported_dtype_encodings
        )
    )
    def test_read_dtype_torch(
        self, pre_buffer, use_threads, use_memory_map, dtype_encoding
    ):

        for n_row_groups, n_columns, chunk_size in [
            (1, 1, 1),
            (2, 2, 1),
            (1, 1, 2),
            (1, 1, 10),
            (1, 1, 100),
            (1, 1, 1_000),
            (1, 1, 10_000),
            (1, 1, 100_000),
            (1, 1, 1_000_000),
            (1, 1, 1_000_001),
        ]:

            dtype = dtype_encoding[0]
            encoding = dtype_encoding[1]

            with self.subTest((n_row_groups, n_columns, chunk_size, dtype, encoding)):
                n_rows = n_row_groups * chunk_size

                with tempfile.TemporaryDirectory() as tmpdirname:
                    path = os.path.join(tmpdirname, "my.parquet")
                    table = get_table(
                        n_rows=n_rows, n_columns=n_columns, data_type=dtype
                    )

                    use_dictionary = False
                    column_encoding = encoding
                    if encoding in ["RLE_DICTIONARY", "PLAIN_DICTIONARY"]:
                        use_dictionary = True
                        column_encoding = None

                    pq.write_table(
                        table,
                        path,
                        row_group_size=chunk_size,
                        use_dictionary=use_dictionary,
                        write_statistics=False,
                        store_schema=False,
                        column_encoding=column_encoding,
                    )

                    tensor = torch.zeros(
                        n_columns,
                        n_rows,
                        dtype=numpy_to_torch_dtype_dict[dtype.to_pandas_dtype()],
                    ).transpose(0, 1)

                    jj.read_into_torch(
                        source=path,
                        metadata=None,
                        tensor=tensor,
                        row_group_indices=range(n_row_groups),
                        column_indices=range(n_columns),
                        pre_buffer=pre_buffer,
                        use_threads=use_threads,
                        use_memory_map=use_memory_map,
                    )

                    pr = pq.ParquetReader()
                    pr.open(path)
                    expected_data = pr.read_all().to_pandas().to_numpy()
                    self.assertTrue(
                        np.array_equal(tensor.numpy(), expected_data),
                        f"{tensor.numpy()}\n{expected_data}",
                    )
                    pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64(), pa.int32(), pa.int64()],
        )
    )
    def test_read_numpy_column_names(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_names=[f"column_{i}" for i in range(n_columns)],
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all().to_pandas().to_numpy()

            self.assertTrue(
                np.array_equal(np_array, expected_data), f"{np_array}\n{expected_data}"
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64(), pa.int32(), pa.int64()],
        )
    )
    def test_read_torch_column_names(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            tensor = torch.zeros(
                n_columns,
                n_rows,
                dtype=numpy_to_torch_dtype_dict[dtype.to_pandas_dtype()],
            ).transpose(0, 1)

            jj.read_into_torch(
                source=path,
                metadata=None,
                tensor=tensor,
                row_group_indices=range(n_row_groups),
                column_names=[f"column_{i}" for i in range(n_columns)],
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all().to_pandas().to_numpy()
            self.assertTrue(
                np.array_equal(tensor.numpy(), expected_data),
                f"{tensor.numpy()}\n{expected_data}",
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_invalid_column(self, pre_buffer, use_threads, use_memory_map, dtype):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_names=[f"foo_bar_{i}" for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"Column 'foo_bar_0' was not found!" in str(context.exception),
                context.exception,
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_indices=[i + 1 for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            ss = [
                f"The file only has {n_columns} columns, requested metadata for column: {n_columns}",
                f"Trying to read column index {n_columns} but row group metadata has only {n_columns} columns",
            ]
            self.assertTrue(
                any(s in str(context.exception) for s in ss), context.exception
            )

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_filesystem(self, pre_buffer, use_threads, use_memory_map, dtype):

        if os.environ.get("JJ_READER_BACKEND") != None:
            self.skipTest(
                "io_uring is enabled but this test is not compatible with io_uring"
            )

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with fs.LocalFileSystem().open_input_file(path) as f:
                jj.read_into_numpy(
                    source=f,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_names=[f"column_{i}" for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all().to_pandas().to_numpy()
            self.assertTrue(
                np.array_equal(np_array, expected_data), f"{np_array}\n{expected_data}"
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_invalid_row_group(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[n_row_groups],
                    column_names=[f"column_{i}" for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            ss = [
                f"The file only has {n_row_groups} row groups, requested metadata for row group: {n_row_groups}",
                f"Trying to read row group {n_row_groups} but file only has {n_row_groups} row groups",
            ]
            self.assertTrue(
                any(s in str(context.exception) for s in ss), context.exception
            )

    @parameterized.expand(
        itertools.product(
            [False, True], [False, True], [False, True], supported_dtype_encodings
        )
    )
    def test_read_data_with_nulls(
        self, pre_buffer, use_threads, use_memory_map, dtype_encoding
    ):

        dtype = dtype_encoding[0]
        encoding = dtype_encoding[1]

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table_with_nulls(
                n_rows=n_rows, n_columns=n_columns, data_type=dtype
            )

            use_dictionary = False
            column_encoding = encoding
            if encoding in ["RLE_DICTIONARY", "PLAIN_DICTIONARY"]:
                use_dictionary = True
                column_encoding = None

            # Convert to PyArrow table
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=use_dictionary,
                write_statistics=False,
                store_schema=True,
                column_encoding=column_encoding,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:

                pr = pq.ParquetReader()
                pr.open(path)
                all_data = pr.read_all()
                pr.close()
                self.assertTrue(all_data.columns[0].type, dtype)

                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_names=[f"column_{i}" for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"Unexpected end of stream" in str(context.exception), context.exception
            )

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_not_enough_rows(self, pre_buffer, use_threads, use_memory_map, dtype):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows + 1, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_names=[f"column_{i}" for i in range(n_columns)],
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"Expected to read {n_rows + 1} rows, but read only {n_rows}!"
                in str(context.exception),
                context.exception,
            )

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_numpy_column_names_mapping(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)

            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_names={
                    f"column_{i}": n_columns - i - 1 for i in range(n_columns)
                },
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all().to_pandas().to_numpy()
            reversed_expected_data = expected_data[:, ::-1]
            self.assertTrue(
                np.array_equal(np_array, reversed_expected_data),
                f"\n{np_array}\n\n{reversed_expected_data}",
            )

            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            for c in range(n_columns):
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_names={f"column_{c}": n_columns - c - 1},
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                np.array_equal(np_array, reversed_expected_data),
                f"\n{np_array}\n\n{reversed_expected_data}",
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_numpy_column_indices_mapping(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_indices={i: n_columns - i - 1 for i in range(n_columns)},
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            expected_data = pr.read_all().to_pandas().to_numpy()
            reversed_expected_data = expected_data[:, ::-1]
            self.assertTrue(
                np.array_equal(np_array, reversed_expected_data),
                f"\n{np_array}\n\n{reversed_expected_data}",
            )

            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            for c in range(n_columns):
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_indices={c: n_columns - c - 1},
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                np.array_equal(np_array, reversed_expected_data),
                f"\n{np_array}\n\n{reversed_expected_data}",
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_numpy_column_indices_multi_mapping(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            pr = pq.ParquetReader()
            pr.open(path)

            column_data = pr.read_all(column_indices=[3]).to_pandas().to_numpy()
            expected_data = np.repeat(column_data, 3, axis=1)

            # Create an empty array
            np_array = np.zeros((n_rows, 3), dtype=dtype.to_pandas_dtype(), order="F")
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_indices=((3, 0), (3, 1)),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_indices=[[3, 2]],
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            self.assertTrue(
                np.array_equal(np_array, expected_data),
                f"\n{np_array}\n\n{expected_data}",
            )
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_numpy_column_names_multi_mapping(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            pr = pq.ParquetReader()
            pr.open(path)

            column_data = pr.read_all(column_indices=[3]).to_pandas().to_numpy()
            expected_data = np.repeat(column_data, 3, axis=1)

            # Create an empty array
            np_array = np.zeros((n_rows, 3), dtype=dtype.to_pandas_dtype(), order="F")
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_names=(("column_3", 0), ("column_3", 1)),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(n_row_groups),
                column_names=[["column_3", 2]],
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            self.assertTrue(
                np.array_equal(np_array, expected_data),
                f"\n{np_array}\n\n{expected_data}",
            )
            pr.close()

    # over 4GB
    @parameterized.expand(
        [
            (200_000_000, pa.float16()),
            (1_100_000_000, pa.float32()),
            (550_000_000, pa.float64()),
        ]
    )
    def test_read_large_array(self, chunk_size, dtype):

        n_row_groups = 1
        n_columns = 1
        n_rows = n_row_groups * chunk_size

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")

            # Create an array of consecutive float32 numbers with 1 million rows
            data = np.arange(n_rows, dtype=dtype.to_pandas_dtype())

            # Create a PyArrow table with a single column
            table = pa.table([data], names=["c0"])
            data = None

            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )
            table = None

            pr = pq.ParquetReader()
            pr.open(path)

            # Create an empty array
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            jj.read_into_numpy(
                source=path,
                metadata=pr.metadata,
                np_array=np_array,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(n_columns),
            )

            self.assertTrue(np.min(np_array) == 0)
            self.assertTrue(np.max(np_array) == n_rows - 1)
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_not_enough_buffer(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows=n_rows, n_columns=n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            # Create an empty array
            np_array = np.zeros(
                (n_rows - 1, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_indices={0: n_columns},
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"Attempted to read {chunk_size} rows into location [0, {n_columns}], but that is beyond target's boundaries."
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(n_row_groups),
                    column_indices=range(n_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )

            self.assertTrue(
                f"Buffer overrun error: Attempted to read {chunk_size} rows into location [{chunk_size}, {n_columns - 1}], but there was space available for only {chunk_size - 1} rows."
                in str(context.exception),
                context.exception,
            )

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_entire_table_with_slices(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )
            pr = pq.ParquetReader()
            pr.open(path)

            # Create an array of zeros
            expected_data = pr.read_all()
            expected_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            expected_array[:chunk_size] = expected_data[chunk_size:]
            expected_array[chunk_size:] = expected_data[:chunk_size]
            row_ranges = [
                slice(chunk_size, 2 * chunk_size),
                slice(0, 1),
                slice(1, chunk_size),
            ]
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
                row_ranges=row_ranges,
            )

            self.assertTrue(np.array_equal(np_array, expected_array))

            # Create an array of zeros
            expected_data = pr.read_all()
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
                row_ranges=[slice(x, x + 1) for x in range(n_rows)],
            )

            self.assertTrue(np.array_equal(np_array, expected_data))

            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_partial_table_with_slices(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )
            pr = pq.ParquetReader()
            pr.open(path)

            # Create an array of zeros
            expected_data = pr.read_all()
            expected_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            expected_array[:chunk_size] = expected_data[:chunk_size]
            row_ranges = [
                slice(0, chunk_size),
            ]
            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=[0],
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
                row_ranges=row_ranges,
            )

            self.assertTrue(np.array_equal(np_array, expected_array))
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_entire_tensor_with_slices(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )
            pr = pq.ParquetReader()
            pr.open(path)

            # Create an array of zeros
            expected_data = pr.read_all()
            expected_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            expected_array[:chunk_size] = expected_data[chunk_size:]
            expected_array[chunk_size:] = expected_data[:chunk_size]
            row_ranges = [
                slice(chunk_size, 2 * chunk_size),
                slice(0, 1),
                slice(1, chunk_size),
            ]
            tensor = torch.zeros(
                n_columns,
                n_rows,
                dtype=numpy_to_torch_dtype_dict[dtype.to_pandas_dtype()],
            ).transpose(0, 1)

            jj.read_into_torch(
                source=path,
                metadata=None,
                tensor=tensor,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
                row_ranges=row_ranges,
            )

            self.assertTrue(np.array_equal(tensor.numpy(), expected_array))
            pr.close()

    @parameterized.expand(
        itertools.product(
            [False, True],
            [False, True],
            [False, True],
            [pa.float16(), pa.float32(), pa.float64()],
        )
    )
    def test_read_with_slices_error_handling(
        self, pre_buffer, use_threads, use_memory_map, dtype
    ):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns, data_type=dtype)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )
            pr = pq.ParquetReader()
            pr.open(path)

            np_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(pr.metadata.num_row_groups),
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[
                        slice(0, 2 * chunk_size),
                    ],
                )
            self.assertTrue(
                f"Requested to read {2 * chunk_size} rows, but the current row group has only {chunk_size} rows"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(pr.metadata.num_row_groups),
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[
                        slice(0, chunk_size),
                        slice(chunk_size, 2 * chunk_size - 1),
                    ],
                )
            self.assertTrue(
                f"Requested to read {chunk_size - 1} rows, but the current row group has {chunk_size} rows"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(ValueError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[slice(0, chunk_size, 2)],
                )
            self.assertTrue(
                f"Row range 'slice(0, {chunk_size}, 2)' is not contiguous"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(ValueError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[slice(None, chunk_size)],
                )
            self.assertTrue(
                f"Row range 'slice(None, {chunk_size}, None)' is not a valid range"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(ValueError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[slice(chunk_size, None)],
                )
            self.assertTrue(
                f"Row range 'slice({chunk_size}, None, None)' is not a valid range"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(ValueError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[slice(chunk_size, chunk_size - 1)],
                )
            self.assertTrue(
                f"Row range 'slice({chunk_size}, {chunk_size - 1}, None)' is not a valid range"
                in str(context.exception),
                context.exception,
            )

            with self.assertRaises(RuntimeError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=[0],
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                    row_ranges=[slice(0, chunk_size), slice(0, chunk_size)],
                )
            self.assertTrue(
                f"Expected to read 2 row ranges, but read only 1!"
                in str(context.exception),
                context.exception,
            )

            pr.close()

    @parameterized.expand(
        itertools.product(
            [pa.float16(), pa.float32(), pa.float64()],
            [
                (1, 1),
                (5, 6),
                (8, 8),
                (16, 16),
                (32, 32),
                (32, 33),
                (33, 32),
                (64, 64),
                (65, 64),
                (64, 65),
                (100, 200),
                (1000, 2000),
            ],
        )
    )
    def test_copy_to_numpy_row_major(self, dtype, n_rows_n_columns):

        n_rows = n_rows_n_columns[0]
        n_columns = n_rows_n_columns[1]

        src_array = get_table(n_rows, n_columns, data_type=dtype).to_pandas().to_numpy()
        dst_array = np.zeros(
            (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        expected_array = np.zeros(
            (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        np.copyto(expected_array, src_array)
        jj.copy_to_numpy_row_major(
            src_array=src_array, dst_array=dst_array, row_indices=range(n_rows)
        )
        self.assertTrue(
            np.array_equal(expected_array, dst_array),
            f"{expected_array}\n!=\n{dst_array}",
        )

        # Reversed rows
        jj.copy_to_numpy_row_major(
            src_array=src_array,
            dst_array=dst_array,
            row_indices=[n_rows - i - 1 for i in range(n_rows)],
        )
        expected_array = expected_array[::-1, :]
        self.assertTrue(
            np.array_equal(expected_array, dst_array),
            f"{expected_array}\n!=\n{dst_array}",
        )

        # Subsets
        src_view = src_array[1 : (n_rows - 1), 1 : (n_columns - 1)]
        dst_array = np.zeros(
            (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        dst_view = dst_array[1 : (n_rows - 1), 1 : (n_columns - 1)]
        expected_array = np.zeros(
            (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        np.copyto(expected_array, src_array)
        expected_array = expected_array[1 : (n_rows - 1), 1 : (n_columns - 1)]
        jj.copy_to_numpy_row_major(
            src_array=src_view, dst_array=dst_view, row_indices=range(n_rows - 2)
        )
        self.assertTrue(
            np.array_equal(expected_array, dst_view),
            f"{expected_array}\n!=\n{dst_view}",
        )

        # Scattered rows
        dst_array = np.zeros(
            (2 * n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        expected_array = np.zeros(
            (2 * n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
        )
        expected_array[::2] = src_array  # Copy to even indices (0, 2, ...)
        jj.copy_to_numpy_row_major(
            src_array=src_array,
            dst_array=dst_array,
            row_indices=[i * 2 for i in range(n_rows)],
        )
        self.assertTrue(
            np.array_equal(expected_array, dst_array),
            f"{expected_array}\n!=\n{dst_array}",
        )

    @parameterized.expand(
        itertools.product([pa.float16(), pa.float32(), pa.float64()], [1, 2], [1, 2])
    )
    def test_copy_to_torch_row_major(self, dtype, n_rows, n_columns):

        src_tensor = torch.tensor(
            get_table(n_rows, n_columns, data_type=dtype).to_pandas().to_numpy()
        )
        dst_tensor = torch.zeros(
            n_rows, n_columns, dtype=numpy_to_torch_dtype_dict[dtype.to_pandas_dtype()]
        )
        expected_tensor = src_tensor.clone().contiguous()
        jj.copy_to_torch_row_major(
            src_tensor=src_tensor, dst_tensor=dst_tensor, row_indices=range(n_rows)
        )
        self.assertTrue(
            torch.equal(expected_tensor, dst_tensor),
            f"{expected_tensor}\n!=\n{dst_tensor}",
        )

        # Reversed rows
        jj.copy_to_torch_row_major(
            src_tensor=src_tensor,
            dst_tensor=dst_tensor,
            row_indices=[n_rows - i - 1 for i in range(n_rows)],
        )
        expected_tensor = torch.flipud(expected_tensor)
        self.assertTrue(
            torch.equal(expected_tensor, dst_tensor),
            f"{expected_tensor}\n!=\n{dst_tensor}",
        )

    @parameterized.expand(
        itertools.product([pa.float16(), pa.float32(), pa.float64()], [5, 50], [6, 60])
    )
    def test_copy_to_row_major_arg_validation(self, dtype, n_rows, n_columns):

        src_array = get_table(n_rows, n_columns, data_type=dtype).to_pandas().to_numpy()

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows, n_columns + 1), dtype=dtype.to_pandas_dtype(), order="C"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array, dst_array=dst_array, row_indices=range(n_rows)
            )
        self.assertTrue(
            f"src_array.shape[1] != dst_array.shape[1], {n_columns} != {n_columns + 1}"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows - 1, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array, dst_array=dst_array, row_indices=range(n_rows)
            )
        self.assertTrue(
            f"src_array.shape[0] > dst_array.shape[0], {n_rows} > {n_rows - 1}"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="F"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array, dst_array=dst_array, row_indices=range(n_rows)
            )
        self.assertTrue(
            f"Expected destination array in a C (row-major) order"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros((n_rows, n_columns), dtype=np.uint8, order="C")
            jj.copy_to_numpy_row_major(
                src_array=src_array, dst_array=dst_array, row_indices=range(n_rows)
            )
        self.assertTrue(
            f"Source and destination arrays have different datatypes, {src_array.dtype} != uint8"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array, dst_array=dst_array, row_indices=range(n_rows - 1)
            )
        self.assertTrue(
            f"Unexpected len of row indices, {n_rows - 1} != {n_rows}"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array,
                dst_array=dst_array,
                row_indices=[i - 1 for i in range(n_rows)],
            )
        self.assertTrue(
            f"Row index = -1 is not in the expected range [0, {n_rows})"
            in str(context.exception),
            context.exception,
        )

        with self.assertRaises(AssertionError) as context:
            dst_array = np.zeros(
                (n_rows, n_columns), dtype=dtype.to_pandas_dtype(), order="C"
            )
            jj.copy_to_numpy_row_major(
                src_array=src_array,
                dst_array=dst_array,
                row_indices=[i + 1 for i in range(n_rows)],
            )
        self.assertTrue(
            f"Row index = {n_rows} is not in the expected range [0, {n_rows})"
            in str(context.exception),
            context.exception,
        )

    @parameterized.expand(
        itertools.product([False, True], [False, True], [False, True])
    )
    def test_read_backends(self, pre_buffer, use_threads, use_memory_map):

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table(n_rows, n_columns)
            pq.write_table(
                table,
                path,
                row_group_size=chunk_size,
                use_dictionary=False,
                write_statistics=False,
                store_schema=False,
            )

            pr = pq.ParquetReader()
            pr.open(path)
            np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")

            jj_backend = os.environ.pop("JJ_READER_BACKEND", None)
            jj.read_into_numpy(
                source=path,
                metadata=None,
                np_array=np_array,
                row_group_indices=range(pr.metadata.num_row_groups),
                column_indices=range(pr.metadata.num_columns),
                pre_buffer=pre_buffer,
                use_threads=use_threads,
                use_memory_map=use_memory_map,
            )

            expected_data = pr.read_all()
            self.assertTrue(np.array_equal(np_array, expected_data))

            os.environ["JJ_READER_BACKEND"] = "foo_bar"
            with self.assertRaises(ValueError) as context:
                jj.read_into_numpy(
                    source=path,
                    metadata=None,
                    np_array=np_array,
                    row_group_indices=range(pr.metadata.num_row_groups),
                    column_indices=range(pr.metadata.num_columns),
                    pre_buffer=pre_buffer,
                    use_threads=use_threads,
                    use_memory_map=use_memory_map,
                )
            self.assertTrue(
                f"Unsupported JJ_READER_BACKEND=foo_bar" in str(context.exception),
                context.exception,
            )
            pr.close()

            os.environ.pop("JJ_READER_BACKEND", None)
            if jj_backend is not None:
                os.environ["JJ_READER_BACKEND"] = jj_backend


if __name__ == "__main__":
    unittest.main()
    # unittest.main(argv=['first-arg-is-ignored', '-k', 'TestJollyJack.test_read_unsupported_encoding_delta_byte_array'])



import jollyjack as jj

def read_into_torch(
    source,
    metadata,
    tensor,
    row_group_indices,
    column_indices=[],
    column_names=[],
    pre_buffer=False,
    use_threads=True,
    use_memory_map=False,
    cache_options=None,
):
    """
    Read parquet data directly into a tensor.

    Parameters
    ----------
    source : str, pathlib.Path, pyarrow.NativeFile, or file-like object
    metadata : FileMetaData, optional
    tensor : The tensor to read into. The shape of the tensor needs to match the number of rows and columns to be read.
    row_group_indices : list[int]
    row_ranges : list[slice], optional
        Specifies slices of destination rows to read into. Each slice defines a range
        of rows in the destination tensor where data should be written.
        Example: [slice(0, 100), slice(200, 300)] will read data into rows 0-99 and 200-299.
        If None, reads into all rows sequentially.
    column_indices : list[int] | dict[int, int] | Iterable[tuple[int, int]], optional
        Specifies the columns to read from the parquet file. Can be:
        - A list of column indices to read.
        - A dict mapping source column indices to target column indices in the tensor.
        - An iterable of tuples, where each tuple contains (source_index, target_index).
    column_names : list[str] | dict[str, int] | Iterable[tuple[str, int]], optional
        Specifies the columns to read from the parquet file by name. Can be:
        - A list of column names to read.
        - A dict mapping source column names to target column indices in the tensor.
        - An iterable of tuples, where each tuple contains (column_name, target_index).
    pre_buffer : bool, default False
    use_threads : bool, default True
    use_memory_map : bool, default False
    cache_options : arrow::io::CacheOptions, default None -> CCacheOptions.LazyDefaults()

    Notes:
    -----
    Either column_indices or column_names must be provided, but not both.
    When using an iterable of tuples for column_indices or column_names,
    each tuple should contain exactly two elements: the source column (index or name)
    and the target column index in the tensor.
    """

    jj._read_into_torch(
        source,
        metadata,
        tensor,
        row_group_indices,
        column_indices,
        column_names,
        pre_buffer,
        use_threads,
        use_memory_map,
    )
    return

def read_into_numpy(
    source,
    metadata,
    np_array,
    row_group_indices,
    row_ranges=[],
    column_indices=[],
    column_names=[],
    pre_buffer=False,
    use_threads=True,
    use_memory_map=False,
    cache_options=None,
):
    """
    Read parquet data directly into a numpy array.
    NumPy array needs to be in a Fortran-style (column-major) order.

    Parameters
    ----------
    source : str, pathlib.Path, pyarrow.NativeFile, or file-like object
    metadata : FileMetaData, optional
    np_array : The array to read into. The shape of the array needs to match the number of rows and columns to be read.
    row_group_indices : list[int]
    row_ranges : list[slice], optional
        Specifies slices of destination rows to read into. Each slice defines a range
        of rows in the destination array where data should be written.
        Example: [slice(0, 100), slice(200, 300)] will read data into rows 0-99 and 200-299.
        If None, reads into all rows sequentially.
    column_indices : list[int] | dict[int, int] | Iterable[tuple[int, int]], optional
        Specifies the columns to read from the parquet file. Can be:
        - A list of column indices to read.
        - A dict mapping source column indices to target column indices in the array.
        - An iterable of tuples, where each tuple contains (source_index, target_index).
    column_names : list[str] | dict[str, int] | Iterable[tuple[str, int]], optional
        Specifies the columns to read from the parquet file by name. Can be:
        - A list of column names to read.
        - A dict mapping source column names to target column indices in the array.
        - An iterable of tuples, where each tuple contains (column_name, target_index).
    pre_buffer : bool, default False
    use_threads : bool, default True
    use_memory_map : bool, default False
    cache_options : pa.CacheOptions(), default None -> CCacheOptions.LazyDefaults()

    Notes:
    -----
    Either column_indices or column_names must be provided, but not both.
    When using an iterable of tuples for column_indices or column_names,
    each tuple should contain exactly two elements: the source column (index or name)
    and the target column index in the numpy array.
    """

    jj._read_into_numpy(
        source,
        metadata,
        np_array,
        row_group_indices,
        column_indices,
        column_names,
        pre_buffer,
        use_threads,
        use_memory_map,
        cache_options,
    )
    return

def copy_to_torch_row_major(src_tensor, dst_tensor, row_indices):
    """
    Copy source column-major tensor to a row-major tensor and shuffle its rows according to provided indices.

    Args:
        src_tensor (torch.Tensor): Source column-major tensor to be copied and shuffled.
        dst_tensor (torch.Tensor): Destination row-major tensor to store the result.
        row_indices (numpy.ndarray): Array of indices specifying the row permutation.

    Raises:
        AssertionError: If tensor shapes do not match or row_indices is invalid.
        RuntimeError: If row_indices has an invalid index.

    Example:
        >>> src = torch.tensor([[1, 2], [3, 4]], dtype=torch.int32)
        >>> dst = torch.zeros(2, 2, dtype=torch.int32)
        >>> indices = np.array([1, 0])
        >>> jj.copy_to_torch_row_major(src, dst, indices)
    """
    return

def copy_to_numpy_row_major(src_array, dst_array, row_indices):
    """
    Copy source column-major array to a row-major array and shuffle its rows according to provided indices.

    Args:
        src_array (numpy.ndarray): Source column-major array to be copied and shuffled.
        dst_array (numpy.ndarray): Destination row-major array to store the result.
        row_indices (numpy.ndarray): Array of indices specifying the row permutation.

    Raises:
        AssertionError: If array shapes do not match or row_indices is invalid.
        RuntimeError: If row_indices has an invalid index.

    Example:
        >>> src = np.array([[1, 2], [3, 4]], dtype=int, order='F')
        >>> dst = np.zeros((2, 2), dtype=int, order='C')
        >>> indices = np.array([1, 0])
        >>> jj.copy_to_numpy_row_major(src, dst, indices)
    """
    return



# Importing pyarrow is necessary to load the runtime libraries
import pyarrow
import pyarrow.parquet

from .package_metadata import __version__, __dependencies__

try:
    from .jollyjack_cython import *
except ImportError as e:
    if any(x in str(e) for x in ["arrow", "parquet"]):
        pyarrow_req = next((r for r in __dependencies__ if r.startswith("pyarrow")), "")
        raise ImportError(
            f"This version of {__package__}={__version__} is built against {pyarrow_req}, please ensure you have it installed. Current pyarrow version is {pyarrow.__version__}. ({str(e)})"
        ) from None
    else:
        raise



# distutils: include_dirs = .

import os
import cython
import pyarrow.parquet as pq
cimport numpy as cnp

from cython.cimports.jollyjack import cjollyjack

from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.vector cimport vector
from libcpp cimport bool
from libc.stdint cimport uint32_t
from pyarrow._parquet cimport *
from pyarrow.lib cimport (CacheOptions, NativeFile, get_reader)
from collections.abc import Iterable

def is_iterable_of_iterables(obj):
    return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)

cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):

    import torch

    read_into_numpy (source = source
        , metadata = metadata
        , np_array = tensor.numpy()
        , row_group_indices = row_group_indices
        , row_ranges = row_ranges
        , column_indices = column_indices
        , column_names = column_names
        , pre_buffer = pre_buffer
        , use_threads = use_threads
        , use_memory_map = use_memory_map
        , cache_options = cache_options
    )

    return

cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):

    cdef vector[int] crow_group_indices = row_group_indices
    cdef vector[int] ccolumn_indices
    cdef uint64_t cstride0_size = np_array.strides[0]
    cdef uint64_t cstride1_size = np_array.strides[1]
    cdef void* cdata = np_array.data
    cdef bool cpre_buffer = pre_buffer
    cdef bool cuse_threads = use_threads
    cdef vector[string] ccolumn_names
    cdef uint64_t cbuffer_size = (np_array.shape[0]) * cstride0_size + (np_array.shape[1] - 1) * cstride1_size
    cdef shared_ptr[CFileMetaData] c_metadata
    cdef vector[int] ctarget_column_indices
    cdef vector[int64_t] ctarget_row_ranges

    if metadata is not None:
        c_metadata = metadata.sp_metadata

    if column_indices and isinstance(column_indices, dict):
        ccolumn_indices = column_indices.keys()
        ctarget_column_indices = column_indices.values()
    elif column_indices and is_iterable_of_iterables(column_indices):
        ccolumn_indices = [item[0] for item in column_indices]
        ctarget_column_indices = [item[1] for item in column_indices]
    elif column_indices:
        assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
        ccolumn_indices = column_indices

    if column_names and isinstance(column_names, dict):
        ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
        ctarget_column_indices = column_names.values()
    elif column_names and is_iterable_of_iterables(column_names):
        ccolumn_names = [item[0].encode('utf8') for item in column_names]
        ctarget_column_indices = [item[1] for item in column_names]
    elif column_names:
        assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
        ccolumn_names = [c.encode('utf8') for c in column_names]

    if row_ranges:
        for sl in row_ranges:
            if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
                raise ValueError(f"Row range '{sl}' is not a valid range")

            if sl.step is not None and sl.step != 1:
                raise ValueError(f"Row range '{sl}' is not contiguous")

        ctarget_row_ranges = [x for sl in row_ranges for x in (sl.start, sl.stop)]

    # Ensure that only one input is set
    assert (column_indices or column_names) and (not column_indices or not column_names), f"Either column_indices or column_names needs to be set"

    # Ensure the input is a 2D array, Fortran order
    assert np_array.ndim == 2, f"Unexpected np_array.ndim, {np_array.ndim} != 2"
    assert np_array.strides[0] <= np_array.strides[1], f"Expected array in a Fortran (column-major) order"

    cdef int64_t cexpected_rows = np_array.shape[0]

    cdef CCacheOptions c_cache_options
    if cache_options is not None:
        c_cache_options = cache_options.unwrap()
    else:
        c_cache_options = CCacheOptions.LazyDefaults()

    cdef shared_ptr[CRandomAccessFile] rd_handle
    cdef c_string pathstr
    cdef bool cuse_o_direct
    # Please note that the `JJ_READER_BACKEND` variable is experimental and may be changed or removed in future versions
    jj_reader_backend = os.environ.get("JJ_READER_BACKEND")
    if jj_reader_backend is None:
        get_reader(source, use_memory_map, &rd_handle)
    elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
        pathstr = source.encode("utf-8")
        cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
        with nogil:
            cjollyjack.ReadIntoMemoryIOUring (pathstr
                , c_metadata
                , np_array.data
                , cbuffer_size
                , cstride0_size
                , cstride1_size
                , ccolumn_indices
                , crow_group_indices
                , ctarget_row_ranges
                , ccolumn_names
                , ctarget_column_indices
                , cpre_buffer
                , cuse_threads
                , cuse_o_direct
                , cexpected_rows
                , c_cache_options)
            return
    else:
        raise ValueError(f"Unsupported JJ_READER_BACKEND={jj_reader_backend}")

    with nogil:
        cjollyjack.ReadIntoMemory (rd_handle
            , c_metadata
            , np_array.data
            , cbuffer_size
            , cstride0_size
            , cstride1_size
            , ccolumn_indices
            , crow_group_indices
            , ctarget_row_ranges
            , ccolumn_names
            , ctarget_column_indices
            , cpre_buffer
            , cuse_threads
            , cexpected_rows
            , c_cache_options)
        return

cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):
    import torch

    copy_to_numpy_row_major (src_array = src_tensor.numpy()
        , dst_array = dst_tensor.numpy()
        , row_indices = row_indices
    )

cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):

    if src_array.shape[0] == 0:
        return # nothing to do

    assert src_array.ndim == 2, f"Unexpected src_array.ndim, {src_array.ndim} != 2"
    assert dst_array.ndim == 2, f"Unexpected dst_array.ndim, {dst_array.ndim} != 2"
    assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"
    assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"
    assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"
    assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"
    assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"
    assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"
    assert min(row_indices) >= 0, f"Row index = {min(row_indices)} is not in the expected range [0, {dst_array.shape[0]})" 
    assert max(row_indices) < dst_array.shape[0], f"Row index = {max(row_indices)} is not in the expected range [0, {dst_array.shape[0]})" 

    cdef vector[int] crow_indices = row_indices
    cdef uint64_t csrc_stride0 = src_array.strides[0]
    cdef uint64_t csrc_stride1 = src_array.strides[1]
    cdef uint64_t csrc_rows = src_array.shape[0]
    cdef uint64_t csrc_columns = src_array.shape[1]
    cdef uint64_t cdst_stride0 = dst_array.strides[0]
    cdef uint64_t cdst_stride1 = dst_array.strides[1]

    with nogil:
        cjollyjack.CopyToRowMajor (src_array.data
            , csrc_stride0
            , csrc_stride1
            , csrc_rows
            , csrc_columns
            , dst_array.data
            , cdst_stride0
            , cdst_stride1
            , crow_indices);
        return



# ```

## How to use:

### Generating a sample parquet file:
# ```
import jollyjack as jj
import pyarrow.parquet as pq
import pyarrow as pa
import numpy as np
import pytest
import os
import unittest
from pyarrow import fs

if os.environ.get("JJ_READER_BACKEND") != None:
    pytest.skip(
        "io_uring is enabled but this test is not compatible with io_uring",
        allow_module_level=True,
    )


chunk_size = 3
n_row_groups = 2
n_columns = 5
n_rows = n_row_groups * chunk_size
path = "my.parquet"

data = np.random.rand(n_rows, n_columns).astype(np.float32)
pa_arrays = [pa.array(data[:, i]) for i in range(n_columns)]
schema = pa.schema([(f"column_{i}", pa.float32()) for i in range(n_columns)])
table = pa.Table.from_arrays(pa_arrays, schema=schema)
pq.write_table(
    table,
    path,
    row_group_size=chunk_size,
    use_dictionary=False,
    write_statistics=True,
    store_schema=False,
    write_page_index=True,
)
# ```

### Generating a numpy array to read into:
# ```
# Create an array of zeros
np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")
# ```

### Reading entire file into numpy array:
# ```
pr = pq.ParquetReader()
pr.open(path)

row_begin = 0
row_end = 0

for rg in range(pr.metadata.num_row_groups):
    row_begin = row_end
    row_end = row_begin + pr.metadata.row_group(rg).num_rows

    # To define which subset of the numpy array we want read into,
    # we need to create a view which shares underlying memory with the target numpy array
    subset_view = np_array[row_begin:row_end, :]
    jj.read_into_numpy(
        source=path,
        metadata=pr.metadata,
        np_array=subset_view,
        row_group_indices=[rg],
        column_indices=range(pr.metadata.num_columns),
    )

# Alternatively
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=range(pr.metadata.num_row_groups),
        column_indices=range(pr.metadata.num_columns),
    )

### Reading columns in reversed order:
# ```
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=range(pr.metadata.num_row_groups),
        column_indices={
            i: pr.metadata.num_columns - i - 1 for i in range(pr.metadata.num_columns)
        },
    )
# ```

### Reading column 3 into multiple destination columns
# ```
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=range(pr.metadata.num_row_groups),
        column_indices=((3, 0), (3, 1)),
    )
# ```

### Sparse reading
# ```
np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=[0],
        row_ranges=[slice(0, 1), slice(4, 6)],
        column_indices=range(pr.metadata.num_columns),
    )
print(np_array)
# ```

### Using cache options
# ```
np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")
cache_options = pa.CacheOptions(hole_size_limit=1024, range_size_limit=2048, lazy=True)
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=[0],
        row_ranges=[slice(0, 1), slice(4, 6)],
        column_indices=range(pr.metadata.num_columns),
        cache_options=cache_options,
        pre_buffer=True,
    )
print(np_array)
# ```

### Using cache options
# ```
np_array = np.zeros((n_rows, n_columns), dtype="f", order="F")
cache_options = pa.CacheOptions(hole_size_limit=1024, range_size_limit=2048, lazy=True)
with fs.LocalFileSystem().open_input_file(path) as f:
    jj.read_into_numpy(
        source=f,
        metadata=None,
        np_array=np_array,
        row_group_indices=[0],
        row_ranges=[slice(0, 1), slice(4, 6)],
        column_indices=range(pr.metadata.num_columns),
        cache_options=cache_options,
        pre_buffer=True,
    )
print(np_array)
# ```

### Generating a torch tensor to read into:
# ```
import torch

# Create a tesnsor and transpose it to get Fortran-style order
tensor = torch.zeros(n_columns, n_rows, dtype=torch.float32).transpose(0, 1)
# ```

### Reading entire file into the tensor:
# ```
pr = pq.ParquetReader()
pr.open(path)

jj.read_into_torch(
    source=path,
    metadata=pr.metadata,
    tensor=tensor,
    row_group_indices=range(pr.metadata.num_row_groups),
    column_indices=range(pr.metadata.num_columns),
    pre_buffer=True,
    use_threads=True,
)

print(tensor)
# ```



#!/usr/bin/env python
import os
import sys

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from distutils.extension import Extension
from Cython.Build import cythonize
import pyarrow
import numpy

include_dirs = [pyarrow.get_include(), numpy.get_include()]
library_dirs = pyarrow.get_library_dirs()

print("include_dirs=", include_dirs)
print("library_dirs=", library_dirs)

extra_compile_args = []
extra_link_args = []
debug = (False,)
files = ["jollyjack/jollyjack_cython.pyx", "jollyjack/jollyjack.cc"]
libraries = ["arrow", "parquet"]
cython_compile_time_env = {}

if os.getenv("DEBUG", "") == "ON":
    print("Building with DEBUG information!")
    extra_compile_args.extend(["-O0", "-DDEBUG"])
    extra_link_args.extend(["-debug:full"])
    debug = True

if os.getenv("ASAN", "") == "ON":
    print("Building with ASan!")
    extra_compile_args.extend(["-fsanitize=address", "-O1", "-fno-omit-frame-pointer"])
    extra_link_args.extend(["-fsanitize=address"])

if sys.platform == "linux":
    files.extend(
        [
            "jollyjack/read_into_memory_io_uring.cc",
            "jollyjack/io_uring_reader_1.cc",
            "jollyjack/direct_reader.cc",
        ]
    )
    libraries.extend(["uring"])
    extra_compile_args.extend(["-DWITH_IO_URING"])

# Define your extension
extensions = [
    Extension(
        "jollyjack.jollyjack_cython",
        files,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        language="c++",
        extra_compile_args=extra_compile_args
        + (["/std:c++20"] if sys.platform.startswith("win") else ["-std=c++20"]),
        extra_link_args=extra_link_args,
    )
]

compiler_directives = {"language_level": 3, "embedsignature": True}
extensions = cythonize(
    extensions,
    compiler_directives=compiler_directives,
    gdb_debug=debug,
    emit_linenums=debug,
)

# Make default named pyarrow shared libs available.
pyarrow.create_library_symlinks()


# Custom build command to dynamically generate metadata file
class GenerateMetadata(build_py):
    def run(self):
        # Get the distribution object
        dist = self.distribution

        package_name = dist.get_name()
        package_version = dist.get_version()
        package_dependencies = dist.install_requires

        print(f"package_name = {package_name}")
        print(f"package_version = {package_version}")
        print(f"package_dependencies = {package_dependencies}")

        output_dir = os.path.join(package_name)
        os.makedirs(output_dir, exist_ok=True)
        metadata_file = os.path.join(output_dir, "package_metadata.py")

        # Write metadata to the file
        with open(metadata_file, "w") as f:
            f.write("# Auto-generated package metadata\n")
            f.write(f"__package__ = '{package_name}'\n")
            f.write(f"__version__ = '{package_version}'\n")
            f.write(f"__dependencies__ = {package_dependencies}\n")

        print(f"Generated metadata file: {metadata_file}")
        print(os.listdir(output_dir))

        # Call the original build_py command
        super().run()


setup(
    packages=["jollyjack"],
    package_dir={"": "."},
    zip_safe=False,
    ext_modules=extensions,
    project_urls={
        "Documentation": "https://github.com/G-Research/JollyJack",
        "Source": "https://github.com/G-Research/JollyJack",
    },
    cmdclass={
        "build_py": GenerateMetadata,  # Use the custom build command
    },
)
